@@ -1,5 +1,5 @@
 use crate::global::handle_error;
-use crate::trace::{NoopTracerProvider, TraceResult};
+use crate::trace::{noop::NoopTracerProvider, SpanContext, StatusCode, TraceResult};
 use crate::{trace, trace::TracerProvider, Context, KeyValue};
 use std::borrow::Cow;
 use std::fmt;
@@ -7,14 +7,159 @@ use std::mem;
 use std::sync::{Arc, RwLock};
 use std::time::SystemTime;
 
+pub trait ObjectSafeSpan {
+    /// An API to record events at a specific time in the context of a given `Span`.
+    ///
+    /// Events SHOULD preserve the order in which they're set. This will typically match
+    /// the ordering of the events' timestamps.
+    ///
+    /// Note that the OpenTelemetry project documents certain ["standard event names and
+    /// keys"](https://github.com/open-telemetry/opentelemetry-specification/tree/v0.5.0/specification/trace/semantic_conventions/README.md)
+    /// which have prescribed semantic meanings.
+    fn add_event_with_timestamp(
+        &mut self,
+        name: Cow<'static, str>,
+        timestamp: SystemTime,
+        attributes: Vec<KeyValue>,
+    );
+
+    /// Returns the `SpanContext` for the given `Span`. The returned value may be used even after
+    /// the `Span is finished. The returned value MUST be the same for the entire `Span` lifetime.
+    fn span_context(&self) -> &SpanContext;
+
+    /// Returns true if this `Span` is recording information like events with the `add_event`
+    /// operation, attributes using `set_attributes`, status with `set_status`, etc.
+    ///
+    /// This flag SHOULD be used to avoid expensive computations of a `Span` attributes or events in
+    /// case when a `Span` is definitely not recorded. Note that any child span's recording is
+    /// determined independently from the value of this flag (typically based on the sampled flag of
+    /// a `TraceFlag` on `SpanContext`).
+    ///
+    /// This flag may be true despite the entire trace being sampled out. This allows to record and
+    /// process information about the individual Span without sending it to the backend. An example
+    /// of this scenario may be recording and processing of all incoming requests for the processing
+    /// and building of SLA/SLO latency charts while sending only a subset - sampled spans - to the
+    /// backend. See also the sampling section of SDK design.
+    ///
+    /// Users of the API should only access the `is_recording` property when instrumenting code and
+    /// never access `SampledFlag` unless used in context propagators.
+    fn is_recording(&self) -> bool;
+
+    /// An API to set a single `Attribute` where the attribute properties are passed
+    /// as arguments. To avoid extra allocations some implementations may offer a separate API for
+    /// each of the possible value types.
+    ///
+    /// An `Attribute` is defined as a `KeyValue` pair.
+    ///
+    /// Attributes SHOULD preserve the order in which they're set. Setting an attribute
+    /// with the same key as an existing attribute SHOULD overwrite the existing
+    /// attribute's value.
+    ///
+    /// Note that the OpenTelemetry project documents certain ["standard
+    /// attributes"](https://github.com/open-telemetry/opentelemetry-specification/tree/v0.5.0/specification/trace/semantic_conventions/README.md)
+    /// that have prescribed semantic meanings.
+    fn set_attribute(&mut self, attribute: KeyValue);
+
+    /// Sets the status of the `Span`. `message` MUST be ignored when the status is `OK` or
+    /// `Unset`.
+    ///
+    /// The order of status is `Ok` > `Error` > `Unset`. That's means set the status
+    /// to `Unset` will always be ignore, set the status to `Error` only works when current
+    /// status is `Unset`, set the status to `Ok` will be consider final and any further call
+    /// to this function will be ignore.
+    fn set_status(&mut self, code: StatusCode, message: String);
+
+    /// Updates the `Span`'s name. After this update, any sampling behavior based on the
+    /// name will depend on the implementation.
+    ///
+    /// It is highly discouraged to update the name of a `Span` after its creation.
+    /// `Span` name is often used to group, filter and identify the logical groups of
+    /// spans. Often, filtering logic will be implemented before the `Span` creation
+    /// for performance reasons, and the name update may interfere with this logic.
+    ///
+    /// The method name is called `update_name` to differentiate this method from the
+    /// regular property. It emphasizes that this operation signifies a
+    /// major change for a `Span` and may lead to re-calculation of sampling or
+    /// filtering decisions made previously depending on the implementation.
+    fn update_name(&mut self, new_name: Cow<'static, str>);
+
+    /// Finishes the `Span`.
+    ///
+    /// Implementations MUST ignore all subsequent calls to `end` (there might be
+    /// exceptions when the tracer is streaming events and has no mutable state
+    /// associated with the Span).
+    ///
+    /// Calls to `end` a Span MUST not have any effects on child `Span`s as they may
+    /// still be running and can be ended later.
+    ///
+    /// This API MUST be non-blocking.
+    fn end(&mut self) {
+        self.end_with_timestamp(crate::time::now());
+    }
+
+    /// Finishes the `Span` with given timestamp
+    ///
+    /// For more details, refer to [`Span::end`]
+    ///
+    /// [`Span::end`]: Span::end()
+    fn end_with_timestamp(&mut self, timestamp: SystemTime);
+}
+
+impl<T: trace::Span> ObjectSafeSpan for T {
+    fn add_event_with_timestamp(
+        &mut self,
+        name: Cow<'static, str>,
+        timestamp: SystemTime,
+        attributes: Vec<KeyValue>,
+    ) {
+        self.add_event_with_timestamp(name, timestamp, attributes)
+    }
+
+    fn span_context(&self) -> &SpanContext {
+        self.span_context()
+    }
+
+    fn is_recording(&self) -> bool {
+        self.is_recording()
+    }
+
+    fn set_attribute(&mut self, attribute: KeyValue) {
+        self.set_attribute(attribute)
+    }
+
+    fn set_status(&mut self, code: StatusCode, message: String) {
+        self.set_status(code, message)
+    }
+
+    fn update_name(&mut self, new_name: Cow<'static, str>) {
+        self.update_name(new_name)
+    }
+
+    fn end_with_timestamp(&mut self, timestamp: SystemTime) {
+        self.end_with_timestamp(timestamp)
+    }
+}
+
 /// Wraps the [`BoxedTracer`]'s [`Span`] so it can be used generically by
 /// applications without knowing the underlying type.
 ///
 /// [`Span`]: crate::trace::Span
-#[derive(Debug)]
-pub struct BoxedSpan(Box<DynSpan>);
+pub struct BoxedSpan(Box<dyn ObjectSafeSpan + Send + Sync>);
 
-type DynSpan = dyn trace::Span + Send + Sync;
+impl BoxedSpan {
+    pub(crate) fn new<T>(span: T) -> Self
+    where
+        T: ObjectSafeSpan + Send + Sync + 'static,
+    {
+        BoxedSpan(Box::new(span))
+    }
+}
+
+impl fmt::Debug for BoxedSpan {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        f.write_str("BoxedSpan")
+    }
+}
 
 impl trace::Span for BoxedSpan {
     /// Records events at a specific time in the context of a given `Span`.
@@ -22,13 +167,16 @@ impl trace::Span for BoxedSpan {
     /// Note that the OpenTelemetry project documents certain ["standard event names and
     /// keys"](https://github.com/open-telemetry/opentelemetry-specification/tree/v0.5.0/specification/trace/semantic_conventions/README.md)
     /// which have prescribed semantic meanings.
-    fn add_event_with_timestamp(
+    fn add_event_with_timestamp<T>(
         &mut self,
-        name: String,
+        name: T,
         timestamp: SystemTime,
         attributes: Vec<KeyValue>,
-    ) {
-        self.0.add_event_with_timestamp(name, timestamp, attributes)
+    ) where
+        T: Into<Cow<'static, str>>,
+    {
+        self.0
+            .add_event_with_timestamp(name.into(), timestamp, attributes)
     }
 
     /// Returns the `SpanContext` for the given `Span`.
@@ -58,8 +206,11 @@ impl trace::Span for BoxedSpan {
     }
 
     /// Updates the `Span`'s name.
-    fn update_name(&mut self, new_name: String) {
-        self.0.update_name(new_name)
+    fn update_name<T>(&mut self, new_name: T)
+    where
+        T: Into<Cow<'static, str>>,
+    {
+        self.0.update_name(new_name.into())
     }
 
     /// Finishes the span with given timestamp.
@@ -73,20 +224,19 @@ impl trace::Span for BoxedSpan {
 ///
 /// [`Tracer`]: crate::trace::Tracer
 /// [`GlobalTracerProvider`]: crate::global::GlobalTracerProvider
-#[derive(Debug)]
-pub struct BoxedTracer(Box<dyn GenericTracer + Send + Sync>);
+pub struct BoxedTracer(Box<dyn ObjectSafeTracer + Send + Sync>);
+
+impl fmt::Debug for BoxedTracer {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        f.write_str("BoxedTracer")
+    }
+}
 
 impl trace::Tracer for BoxedTracer {
     /// Global tracer uses `BoxedSpan`s so that it can be a global singleton,
     /// which is not possible if it takes generic type parameters.
     type Span = BoxedSpan;
 
-    /// Returns a span with an inactive `SpanContext`. Used by functions that
-    /// need to return a default span like `get_active_span` if no span is present.
-    fn invalid(&self) -> Self::Span {
-        BoxedSpan(self.0.invalid_boxed())
-    }
-
     /// Starts a new `Span`.
     ///
     /// Each span has zero or one parent spans and zero or more child spans, which
@@ -94,11 +244,11 @@ impl trace::Tracer for BoxedTracer {
     /// trace. A span is said to be a _root span_ if it does not have a parent. Each
     /// trace includes a single root span, which is the shared ancestor of all other
     /// spans in the trace.
-    fn start_with_context<T>(&self, name: T, cx: Context) -> Self::Span
+    fn start_with_context<T>(&self, name: T, parent_cx: &Context) -> Self::Span
     where
         T: Into<Cow<'static, str>>,
     {
-        BoxedSpan(self.0.start_with_context_boxed(name.into(), cx))
+        BoxedSpan(self.0.start_with_context_boxed(name.into(), parent_cx))
     }
 
     /// Creates a span builder
@@ -112,8 +262,8 @@ impl trace::Tracer for BoxedTracer {
     }
 
     /// Create a span from a `SpanBuilder`
-    fn build(&self, builder: trace::SpanBuilder) -> Self::Span {
-        BoxedSpan(self.0.build_boxed(builder))
+    fn build_with_context(&self, builder: trace::SpanBuilder, parent_cx: &Context) -> Self::Span {
+        BoxedSpan(self.0.build_with_context_boxed(builder, parent_cx))
     }
 }
 
@@ -121,39 +271,47 @@ impl trace::Tracer for BoxedTracer {
 /// instances by mirroring the interface and boxing the return types.
 ///
 /// [`Tracer`]: crate::trace::Tracer
-pub trait GenericTracer: fmt::Debug + 'static {
-    /// Create a new invalid span for use in cases where there are no active spans.
-    fn invalid_boxed(&self) -> Box<DynSpan>;
+pub trait ObjectSafeTracer {
+    /// Returns a trait object so the underlying implementation can be swapped
+    /// out at runtime.
+    fn start_with_context_boxed(
+        &self,
+        name: Cow<'static, str>,
+        parent_cx: &Context,
+    ) -> Box<dyn ObjectSafeSpan + Send + Sync>;
 
     /// Returns a trait object so the underlying implementation can be swapped
     /// out at runtime.
-    fn start_with_context_boxed(&self, name: Cow<'static, str>, cx: Context) -> Box<DynSpan>;
-
-    /// Returns a trait object so the underlying implementation can be swapped
-    /// out at runtime.
-    fn build_boxed(&self, builder: trace::SpanBuilder) -> Box<DynSpan>;
+    fn build_with_context_boxed(
+        &self,
+        builder: trace::SpanBuilder,
+        parent_cx: &Context,
+    ) -> Box<dyn ObjectSafeSpan + Send + Sync>;
 }
 
-impl<S, T> GenericTracer for T
+impl<S, T> ObjectSafeTracer for T
 where
     S: trace::Span + Send + Sync + 'static,
     T: trace::Tracer<Span = S>,
 {
-    /// Create a new invalid span for use in cases where there are no active spans.
-    fn invalid_boxed(&self) -> Box<DynSpan> {
-        Box::new(self.invalid())
+    /// Returns a trait object so the underlying implementation can be swapped
+    /// out at runtime.
+    fn start_with_context_boxed(
+        &self,
+        name: Cow<'static, str>,
+        parent_cx: &Context,
+    ) -> Box<dyn ObjectSafeSpan + Send + Sync> {
+        Box::new(self.start_with_context(name, parent_cx))
     }
 
     /// Returns a trait object so the underlying implementation can be swapped
     /// out at runtime.
-    fn start_with_context_boxed(&self, name: Cow<'static, str>, cx: Context) -> Box<DynSpan> {
-        Box::new(self.start_with_context(name, cx))
-    }
-
-    /// Returns a trait object so the underlying implementation can be swapped
-    /// out at runtime.
-    fn build_boxed(&self, builder: trace::SpanBuilder) -> Box<DynSpan> {
-        Box::new(self.build(builder))
+    fn build_with_context_boxed(
+        &self,
+        builder: trace::SpanBuilder,
+        parent_cx: &Context,
+    ) -> Box<dyn ObjectSafeSpan + Send + Sync> {
+        Box::new(self.build_with_context(builder, parent_cx))
     }
 }
 
@@ -162,31 +320,34 @@ where
 ///
 /// [`TracerProvider`]: crate::trace::TracerProvider
 /// [`GlobalTracerProvider`]: crate::global::GlobalTracerProvider
-pub trait GenericTracerProvider: fmt::Debug + 'static {
-    /// Creates a named tracer instance that is a trait object through the underlying `TracerProvider`.
-    fn get_tracer_boxed(
+pub trait ObjectSafeTracerProvider {
+    /// Creates a versioned named tracer instance that is a trait object through the underlying
+    /// `TracerProvider`.
+    fn versioned_tracer_boxed(
         &self,
-        name: &'static str,
+        name: Cow<'static, str>,
         version: Option<&'static str>,
-    ) -> Box<dyn GenericTracer + Send + Sync>;
+        schema_url: Option<&'static str>,
+    ) -> Box<dyn ObjectSafeTracer + Send + Sync>;
 
     /// Force flush all remaining spans in span processors and return results.
     fn force_flush(&self) -> Vec<TraceResult<()>>;
 }
 
-impl<S, T, P> GenericTracerProvider for P
+impl<S, T, P> ObjectSafeTracerProvider for P
 where
     S: trace::Span + Send + Sync + 'static,
-    T: trace::Tracer<Span = S> + Send + Sync,
+    T: trace::Tracer<Span = S> + Send + Sync + 'static,
     P: trace::TracerProvider<Tracer = T>,
 {
-    /// Return a boxed generic tracer
-    fn get_tracer_boxed(
+    /// Return a versioned boxed tracer
+    fn versioned_tracer_boxed(
         &self,
-        name: &'static str,
+        name: Cow<'static, str>,
         version: Option<&'static str>,
-    ) -> Box<dyn GenericTracer + Send + Sync> {
-        Box::new(self.get_tracer(name, version))
+        schema_url: Option<&'static str>,
+    ) -> Box<dyn ObjectSafeTracer + Send + Sync> {
+        Box::new(self.versioned_tracer(name, version, schema_url))
     }
 
     fn force_flush(&self) -> Vec<TraceResult<()>> {
@@ -199,9 +360,15 @@ where
 /// [`BoxedTracer`] instances.
 ///
 /// [`TracerProvider`]: crate::trace::TracerProvider
-#[derive(Clone, Debug)]
+#[derive(Clone)]
 pub struct GlobalTracerProvider {
-    provider: Arc<dyn GenericTracerProvider + Send + Sync>,
+    provider: Arc<dyn ObjectSafeTracerProvider + Send + Sync>,
+}
+
+impl fmt::Debug for GlobalTracerProvider {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        f.write_str("GlobalTracerProvider")
+    }
 }
 
 impl GlobalTracerProvider {
@@ -209,8 +376,8 @@ impl GlobalTracerProvider {
     fn new<P, T, S>(provider: P) -> Self
     where
         S: trace::Span + Send + Sync + 'static,
-        T: trace::Tracer<Span = S> + Send + Sync,
-        P: trace::TracerProvider<Tracer = T> + Send + Sync,
+        T: trace::Tracer<Span = S> + Send + Sync + 'static,
+        P: trace::TracerProvider<Tracer = T> + Send + Sync + 'static,
     {
         GlobalTracerProvider {
             provider: Arc::new(provider),
@@ -221,9 +388,17 @@ impl GlobalTracerProvider {
 impl trace::TracerProvider for GlobalTracerProvider {
     type Tracer = BoxedTracer;
 
-    /// Find or create a named tracer using the global provider.
-    fn get_tracer(&self, name: &'static str, version: Option<&'static str>) -> Self::Tracer {
-        BoxedTracer(self.provider.get_tracer_boxed(name, version))
+    /// Create a versioned tracer using the global provider.
+    fn versioned_tracer(
+        &self,
+        name: impl Into<Cow<'static, str>>,
+        version: Option<&'static str>,
+        schema_url: Option<&'static str>,
+    ) -> Self::Tracer {
+        BoxedTracer(
+            self.provider
+                .versioned_tracer_boxed(name.into(), version, schema_url),
+        )
     }
 
     /// Force flush all remaining spans in span processors and return results.
@@ -234,7 +409,7 @@ impl trace::TracerProvider for GlobalTracerProvider {
 
 lazy_static::lazy_static! {
     /// The global `Tracer` provider singleton.
-    static ref GLOBAL_TRACER_PROVIDER: RwLock<GlobalTracerProvider> = RwLock::new(GlobalTracerProvider::new(trace::NoopTracerProvider::new()));
+    static ref GLOBAL_TRACER_PROVIDER: RwLock<GlobalTracerProvider> = RwLock::new(GlobalTracerProvider::new(trace::noop::NoopTracerProvider::new()));
 }
 
 /// Returns an instance of the currently configured global [`TracerProvider`] through
@@ -253,21 +428,11 @@ pub fn tracer_provider() -> GlobalTracerProvider {
 ///
 /// If the name is an empty string, the provider will use a default name.
 ///
-/// This is a more convenient way of expressing `global::tracer_provider().get_tracer(name, None)`.
+/// This is a more convenient way of expressing `global::tracer_provider().tracer(name)`.
 ///
 /// [`Tracer`]: crate::trace::Tracer
-pub fn tracer(name: &'static str) -> BoxedTracer {
-    tracer_provider().get_tracer(name, None)
-}
-
-/// Creates a named instance of [`Tracer`] with version info via the configured [`GlobalTracerProvider`]
-///
-/// If the name is an empty string, the provider will use a default name.
-/// If the version is an empty string, it will be used as part of instrumentation library information.
-///
-/// [`Tracer`]: crate::trace::Tracer
-pub fn tracer_with_version(name: &'static str, version: &'static str) -> BoxedTracer {
-    tracer_provider().get_tracer(name, Some(version))
+pub fn tracer(name: impl Into<Cow<'static, str>>) -> BoxedTracer {
+    tracer_provider().tracer(name.into())
 }
 
 /// Sets the given [`TracerProvider`] instance as the current global provider.
@@ -279,8 +444,8 @@ pub fn tracer_with_version(name: &'static str, version: &'static str) -> BoxedTr
 pub fn set_tracer_provider<P, T, S>(new_provider: P) -> GlobalTracerProvider
 where
     S: trace::Span + Send + Sync + 'static,
-    T: trace::Tracer<Span = S> + Send + Sync,
-    P: trace::TracerProvider<Tracer = T> + Send + Sync,
+    T: trace::Tracer<Span = S> + Send + Sync + 'static,
+    P: trace::TracerProvider<Tracer = T> + Send + Sync + 'static,
 {
     let mut tracer_provider = GLOBAL_TRACER_PROVIDER
         .write()
@@ -332,16 +497,9 @@ mod tests {
     use crate::runtime;
     #[cfg(any(feature = "rt-tokio", feature = "rt-tokio-current-thread"))]
     use crate::sdk::trace::TraceRuntime;
-    use crate::trace::NoopTracer;
     #[cfg(any(feature = "rt-tokio", feature = "rt-tokio-current-thread"))]
     use crate::trace::Tracer;
-    use std::{
-        fmt::Debug,
-        io::Write,
-        sync::Mutex,
-        thread::{self, sleep},
-        time::Duration,
-    };
+    use std::{fmt::Debug, io::Write, sync::Mutex};
 
     #[derive(Debug)]
     struct AssertWriter {
@@ -390,99 +548,6 @@ mod tests {
         }
     }
 
-    #[derive(Debug)]
-    struct TestTracerProvider {
-        debug_msg: &'static str,
-    }
-
-    impl Default for TestTracerProvider {
-        fn default() -> Self {
-            TestTracerProvider { debug_msg: "" }
-        }
-    }
-
-    impl TestTracerProvider {
-        fn new(debug_msg: &'static str) -> Self {
-            TestTracerProvider { debug_msg }
-        }
-    }
-
-    impl TracerProvider for TestTracerProvider {
-        type Tracer = NoopTracer;
-
-        fn get_tracer(&self, _name: &'static str, _version: Option<&'static str>) -> Self::Tracer {
-            NoopTracer::default()
-        }
-
-        fn force_flush(&self) -> Vec<TraceResult<()>> {
-            Vec::new()
-        }
-    }
-
-    #[test]
-    #[ignore = "requires --test-threads=1"]
-    fn test_set_tracer_provider() {
-        let _ = set_tracer_provider(TestTracerProvider::new("global one"));
-
-        {
-            let _ = set_tracer_provider(TestTracerProvider::new("inner one"));
-            assert!(format!("{:?}", tracer_provider()).contains("inner one"));
-        }
-
-        assert!(format!("{:?}", tracer_provider()).contains("inner one"));
-    }
-
-    #[test]
-    #[ignore = "requires --test-threads=1"]
-    fn test_set_tracer_provider_in_another_thread() {
-        let _ = set_tracer_provider(TestTracerProvider::new("global one"));
-
-        let handle = thread::spawn(move || {
-            assert!(format!("{:?}", tracer_provider()).contains("global one"));
-        });
-
-        println!("{:?}", tracer_provider());
-
-        let _ = handle.join();
-    }
-
-    #[test]
-    #[ignore = "requires --test-threads=1"]
-    fn test_set_tracer_provider_in_another_function() {
-        let setup = || {
-            let _ = set_tracer_provider(TestTracerProvider::new("global one"));
-            assert!(format!("{:?}", tracer_provider()).contains("global one"))
-        };
-
-        setup();
-
-        assert!(format!("{:?}", tracer_provider()).contains("global one"))
-    }
-
-    #[test]
-    #[ignore = "requires --test-threads=1"]
-    fn test_set_two_provider_in_two_thread() {
-        let (sender, recv) = std::sync::mpsc::channel();
-        let (sender1, sender2) = (sender.clone(), sender);
-        let _handle1 = thread::spawn(move || {
-            sleep(Duration::from_secs(1));
-            let _previous = set_tracer_provider(TestTracerProvider::new("thread 1"));
-            sleep(Duration::from_secs(2));
-            let _ = sender1.send(format!("thread 1: {:?}", tracer_provider()));
-        });
-        let _handle2 = thread::spawn(move || {
-            sleep(Duration::from_secs(2));
-            let _previous = set_tracer_provider(TestTracerProvider::new("thread 2"));
-            sleep(Duration::from_secs(1));
-            let _ = sender2.send(format!("thread 2 :{:?}", tracer_provider()));
-        });
-
-        let first_resp = recv.recv().unwrap();
-        let second_resp = recv.recv().unwrap();
-        assert!(first_resp.contains("thread 2"));
-        assert!(second_resp.contains("thread 2"));
-    }
-
     #[cfg(any(feature = "rt-tokio", feature = "rt-tokio-current-thread"))]
     fn build_batch_tracer_provider<R: TraceRuntime>(
         assert_writer: AssertWriter,
