@@ -14,6 +14,8 @@
 use serde::{Deserialize, Serialize};
 use std::collections::VecDeque;
 use std::fmt;
+use std::hash::Hash;
+use std::num::ParseIntError;
 use std::ops::{BitAnd, BitOr, Not};
 use std::str::FromStr;
 use thiserror::Error;
@@ -93,89 +95,127 @@ impl fmt::LowerHex for TraceFlags {
     }
 }
 
-/// TraceId is an 16-byte value which uniquely identifies a given trace
-/// The actual `u128` value is wrapped in a tuple struct in order to leverage the newtype pattern
+/// A 16-byte value which identifies a given trace.
+///
+/// The id is valid if it contains at least one non-zero byte.
 #[cfg_attr(feature = "serialize", derive(Deserialize, Serialize))]
-#[derive(Clone, Debug, PartialEq, Eq, Copy, Hash)]
-pub struct TraceId(u128);
+#[derive(Clone, PartialEq, Eq, Copy, Hash)]
+pub struct TraceId(pub(crate) u128);
 
 impl TraceId {
-    /// Construct a new invalid (zero-valued) TraceId
-    pub fn invalid() -> Self {
-        TraceId(0)
+    /// Invalid trace id
+    pub const INVALID: TraceId = TraceId(0);
+
+    /// Create a trace id from its representation as a byte array.
+    pub const fn from_bytes(bytes: [u8; 16]) -> Self {
+        TraceId(u128::from_be_bytes(bytes))
     }
 
-    /// Convert from u128 to TraceId
-    pub fn from_u128(item: u128) -> Self {
-        TraceId(item)
-    }
-
-    /// Convert from TraceId to u128
-    pub fn to_u128(self) -> u128 {
-        self.0
-    }
-
-    /// Convert from TraceId to Hexadecimal String
-    pub fn to_hex(self) -> String {
-        format!("{:032x}", self.0)
-    }
-
-    /// Convert from TraceId to Big-Endian byte array
-    pub fn to_byte_array(self) -> [u8; 16] {
+    /// Return the representation of this trace id as a byte array.
+    pub const fn to_bytes(self) -> [u8; 16] {
         self.0.to_be_bytes()
     }
 
-    /// Construct a new TraceId from Hexadecimal String
-    pub fn from_hex(hex: &str) -> Self {
-        TraceId(u128::from_str_radix(hex, 16).unwrap_or(0))
-    }
-
-    /// Construct a new TraceId from Big-Endian byte array
-    pub fn from_byte_array(byte_array: [u8; 16]) -> Self {
-        TraceId(u128::from_be_bytes(byte_array))
+    /// Converts a string in base 16 to a trace id.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use opentelemetry::trace::TraceId;
+    ///
+    /// assert!(TraceId::from_hex("42").is_ok());
+    /// assert!(TraceId::from_hex("58406520a006649127e371903a2de979").is_ok());
+    ///
+    /// assert!(TraceId::from_hex("not_hex").is_err());
+    /// ```
+    pub fn from_hex(hex: &str) -> Result<Self, ParseIntError> {
+        u128::from_str_radix(hex, 16).map(TraceId)
     }
 }
 
-/// SpanId is an 8-byte value which uniquely identifies a given span within a trace
-/// The actual `u64` value is wrapped in a tuple struct in order to leverage the newtype pattern
+impl From<[u8; 16]> for TraceId {
+    fn from(bytes: [u8; 16]) -> Self {
+        TraceId::from_bytes(bytes)
+    }
+}
+
+impl fmt::Debug for TraceId {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        f.write_fmt(format_args!("{:032x}", self.0))
+    }
+}
+
+impl fmt::Display for TraceId {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        f.write_fmt(format_args!("{:032x}", self.0))
+    }
+}
+
+impl fmt::LowerHex for TraceId {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        fmt::LowerHex::fmt(&self.0, f)
+    }
+}
+
+/// An 8-byte value which identifies a given span.
+///
+/// The id is valid if it contains at least one non-zero byte.
 #[cfg_attr(feature = "serialize", derive(Deserialize, Serialize))]
-#[derive(Clone, Debug, PartialEq, Eq, Copy, Hash)]
-pub struct SpanId(u64);
+#[derive(Clone, PartialEq, Eq, Copy, Hash)]
+pub struct SpanId(pub(crate) u64);
 
 impl SpanId {
-    /// Construct a new invalid (zero-valued) SpanId
-    pub fn invalid() -> Self {
-        SpanId(0)
+    /// Invalid span id
+    pub const INVALID: SpanId = SpanId(0);
+
+    /// Create a span id from its representation as a byte array.
+    pub const fn from_bytes(bytes: [u8; 8]) -> Self {
+        SpanId(u64::from_be_bytes(bytes))
     }
 
-    /// Convert from u64 to SpanId
-    pub fn from_u64(item: u64) -> Self {
-        SpanId(item)
-    }
-
-    /// Convert from SpanId to u64
-    pub fn to_u64(self) -> u64 {
-        self.0
-    }
-
-    /// Convert from SpanId to Hexadecimal String
-    pub fn to_hex(self) -> String {
-        format!("{:016x}", self.0)
-    }
-
-    /// Convert from SpanId to Big-Endian byte array
-    pub fn to_byte_array(self) -> [u8; 8] {
+    /// Return the representation of this span id as a byte array.
+    pub const fn to_bytes(self) -> [u8; 8] {
         self.0.to_be_bytes()
     }
 
-    /// Construct a new SpanId from Hexadecimal String
-    pub fn from_hex(hex: &str) -> Self {
-        SpanId(u64::from_str_radix(hex, 16).unwrap_or(0))
+    /// Converts a string in base 16 to a span id.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// use opentelemetry::trace::SpanId;
+    ///
+    /// assert!(SpanId::from_hex("42").is_ok());
+    /// assert!(SpanId::from_hex("58406520a0066491").is_ok());
+    ///
+    /// assert!(SpanId::from_hex("not_hex").is_err());
+    /// ```
+    pub fn from_hex(hex: &str) -> Result<Self, ParseIntError> {
+        u64::from_str_radix(hex, 16).map(SpanId)
     }
+}
 
-    /// Construct a new SpanId from Big-Endian byte array
-    pub fn from_byte_array(byte_array: [u8; 8]) -> Self {
-        SpanId(u64::from_be_bytes(byte_array))
+impl From<[u8; 8]> for SpanId {
+    fn from(bytes: [u8; 8]) -> Self {
+        SpanId::from_bytes(bytes)
+    }
+}
+
+impl fmt::Debug for SpanId {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        f.write_fmt(format_args!("{:016x}", self.0))
+    }
+}
+
+impl fmt::Display for SpanId {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        f.write_fmt(format_args!("{:016x}", self.0))
+    }
+}
+
+impl fmt::LowerHex for SpanId {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        fmt::LowerHex::fmt(&self.0, f)
     }
 }
 
@@ -187,13 +227,13 @@ impl SpanId {
 ///
 /// [W3C specification]: https://www.w3.org/TR/trace-context/#tracestate-header
 #[cfg_attr(feature = "serialize", derive(Deserialize, Serialize))]
-#[derive(Clone, Debug, Default, Eq, PartialEq)]
+#[derive(Clone, Debug, Default, Eq, PartialEq, Hash)]
 pub struct TraceState(Option<VecDeque<(String, String)>>);
 
 impl TraceState {
-    /// Validates that the given `TraceState` list-member key is valid per the [W3 Spec]['spec'].
+    /// Validates that the given `TraceState` list-member key is valid per the [W3 Spec].
     ///
-    /// ['spec']: https://www.w3.org/TR/trace-context/#key
+    /// [W3 Spec]: https://www.w3.org/TR/trace-context/#key
     fn valid_key(key: &str) -> bool {
         if key.len() > 256 {
             return false;
@@ -209,7 +249,7 @@ impl TraceState {
             if i == 0 && (!b.is_ascii_lowercase() && !b.is_ascii_digit()) {
                 return false;
             } else if b == b'@' {
-                if vendor_start.is_some() || i < key.len() - 14 {
+                if vendor_start.is_some() || i + 14 < key.len() {
                     return false;
                 }
                 vendor_start = Some(i);
@@ -223,9 +263,9 @@ impl TraceState {
         true
     }
 
-    /// Validates that the given `TraceState` list-member value is valid per the [W3 Spec]['spec'].
+    /// Validates that the given `TraceState` list-member value is valid per the [W3 Spec].
     ///
-    /// ['spec']: https://www.w3.org/TR/trace-context/#value
+    /// [W3 Spec]: https://www.w3.org/TR/trace-context/#value
     fn valid_value(value: &str) -> bool {
         if value.len() > 256 {
             return false;
@@ -290,10 +330,10 @@ impl TraceState {
 
     /// Inserts the given key-value pair into the `TraceState`. If a value already exists for the
     /// given key, this updates the value and updates the value's position. If the key or value are
-    /// invalid per the [W3 Spec]['spec'] an `Err` is returned, else a new `TraceState` with the
+    /// invalid per the [W3 Spec] an `Err` is returned, else a new `TraceState` with the
     /// updated key/value is returned.
     ///
-    /// ['spec']: https://www.w3.org/TR/trace-context/#list
+    /// [W3 Spec]: https://www.w3.org/TR/trace-context/#mutating-the-tracestate-field
     pub fn insert<K, V>(&self, key: K, value: V) -> Result<TraceState, TraceStateError>
     where
         K: Into<String>,
@@ -316,11 +356,12 @@ impl TraceState {
     }
 
     /// Removes the given key-value pair from the `TraceState`. If the key is invalid per the
-    /// [W3 Spec]['spec'] an `Err` is returned. Else, a new `TraceState`
+    /// [W3 Spec] an `Err` is returned. Else, a new `TraceState`
     /// with the removed entry is returned.
     ///
     /// If the key is not in `TraceState`. The original `TraceState` will be cloned and returned.
-    /// ['spec']: https://www.w3.org/TR/trace-context/#list
+    ///
+    /// [W3 Spec]: https://www.w3.org/TR/trace-context/#mutating-the-tracestate-field
     pub fn delete<K: Into<String>>(&self, key: K) -> Result<TraceState, TraceStateError> {
         let key = key.into();
         if !TraceState::valid_key(key.as_str()) {
@@ -387,15 +428,15 @@ impl FromStr for TraceState {
 #[derive(Error, Debug)]
 #[non_exhaustive]
 pub enum TraceStateError {
-    /// The key is invalid. See https://www.w3.org/TR/trace-context/#key for requirement for keys.
+    /// The key is invalid. See <https://www.w3.org/TR/trace-context/#key> for requirement for keys.
     #[error("{0} is not a valid key in TraceState, see https://www.w3.org/TR/trace-context/#key for more details")]
     InvalidKey(String),
 
-    /// The value is invalid. See https://www.w3.org/TR/trace-context/#value for requirement for values.
+    /// The value is invalid. See <https://www.w3.org/TR/trace-context/#value> for requirement for values.
     #[error("{0} is not a valid value in TraceState, see https://www.w3.org/TR/trace-context/#value for more details")]
     InvalidValue(String),
 
-    /// The value is invalid. See https://www.w3.org/TR/trace-context/#list for requirement for list members.
+    /// The value is invalid. See <https://www.w3.org/TR/trace-context/#list> for requirement for list members.
     #[error("{0} is not a valid list member in TraceState, see https://www.w3.org/TR/trace-context/#list for more details")]
     InvalidList(String),
 }
@@ -405,7 +446,7 @@ pub enum TraceStateError {
 /// Spans that do not have the `sampled` flag set in their [`TraceFlags`] will
 /// be ignored by most tracing tools.
 #[cfg_attr(feature = "serialize", derive(Deserialize, Serialize))]
-#[derive(Clone, Debug, PartialEq)]
+#[derive(Clone, Debug, PartialEq, Hash, Eq)]
 pub struct SpanContext {
     trace_id: TraceId,
     span_id: SpanId,
@@ -418,8 +459,8 @@ impl SpanContext {
     /// Create an invalid empty span context
     pub fn empty_context() -> Self {
         SpanContext::new(
-            TraceId::invalid(),
-            SpanId::invalid(),
+            TraceId::INVALID,
+            SpanId::INVALID,
             TraceFlags::default(),
             false,
             TraceState::default(),
@@ -517,22 +558,24 @@ mod tests {
     #[test]
     fn test_trace_id() {
         for test_case in trace_id_test_data() {
-            assert_eq!(test_case.0.to_hex(), test_case.1);
-            assert_eq!(test_case.0.to_byte_array(), test_case.2);
+            assert_eq!(format!("{}", test_case.0), test_case.1);
+            assert_eq!(format!("{:032x}", test_case.0), test_case.1);
+            assert_eq!(test_case.0.to_bytes(), test_case.2);
 
-            assert_eq!(test_case.0, TraceId::from_hex(test_case.1));
-            assert_eq!(test_case.0, TraceId::from_byte_array(test_case.2));
+            assert_eq!(test_case.0, TraceId::from_hex(test_case.1).unwrap());
+            assert_eq!(test_case.0, TraceId::from_bytes(test_case.2));
         }
     }
 
     #[test]
     fn test_span_id() {
         for test_case in span_id_test_data() {
-            assert_eq!(test_case.0.to_hex(), test_case.1);
-            assert_eq!(test_case.0.to_byte_array(), test_case.2);
+            assert_eq!(format!("{}", test_case.0), test_case.1);
+            assert_eq!(format!("{:016x}", test_case.0), test_case.1);
+            assert_eq!(test_case.0.to_bytes(), test_case.2);
 
-            assert_eq!(test_case.0, SpanId::from_hex(test_case.1));
-            assert_eq!(test_case.0, SpanId::from_byte_array(test_case.2));
+            assert_eq!(test_case.0, SpanId::from_hex(test_case.1).unwrap());
+            assert_eq!(test_case.0, SpanId::from_bytes(test_case.2));
         }
     }
 
@@ -563,6 +606,23 @@ mod tests {
         }
     }
 
+    #[test]
+    fn test_trace_state_key() {
+        let test_data: Vec<(&'static str, bool)> = vec![
+            ("123", true),
+            ("bar", true),
+            ("foo@bar", true),
+            ("foo@0123456789abcdef", false),
+            ("foo@012345678", true),
+            ("FOO@BAR", false),
+            ("你好", false),
+        ];
+
+        for (key, expected) in test_data {
+            assert_eq!(TraceState::valid_key(key), expected, "test key: {:?}", key);
+        }
+    }
+
     #[test]
     fn test_trace_state_insert() {
         let trace_state = TraceState::from_key_value(vec![("foo", "bar")]).unwrap();
